"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import timedelta, date, datetime
from pytz import UTC
from selenium.webdriver.support.ui import Select, WebDriverWait
from selenium.webdriver.common.by import By
from selenium.common.exceptions import TimeoutException
from decimal import Decimal
import requests

from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.models.budget_model import Budget


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransactions(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Transactions - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransactionsDefault(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/transactions')

    def test_table(self, selenium):
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts == [
            [
                (self.dt + timedelta(days=4)).date().strftime('%Y-%m-%d'),
                '$111.13',
                'T1foo',
                'BankOne (1)',
                'Periodic1 (1)',
                'Yes (1)',
                '$111.11',
                '',
                'Yes (1)'
            ],
            [
                self.dt.date().strftime('%Y-%m-%d'),
                '-$333.33',
                'T2',
                'BankTwoStale (2)',
                'Standing1 (4)',
                'Yes (3)',
                '',
                '',
                ''
            ],
            [
                (self.dt - timedelta(days=2)).date().strftime('%Y-%m-%d'),
                '$222.22',
                'T3',
                'CreditOne (3)',
                'Periodic2 (2)',
                '',
                '',
                '$12.34',
                ''
            ],
            [
                (self.dt - timedelta(days=35)).date().strftime('%Y-%m-%d'),
                '$322.32',
                'T4split',
                'CreditOne (3)',
                'Periodic2 (2) ($222.22)\nPeriodic1 (1) ($100.10)',
                '',
                '',
                '$34.56',
                ''
            ]
        ]
        linkcols = [
            [
                c[2].get_attribute('innerHTML'),
                c[3].get_attribute('innerHTML'),
                c[4].get_attribute('innerHTML'),
                c[5].get_attribute('innerHTML'),
                c[8].get_attribute('innerHTML')
            ]
            for c in elems
        ]
        assert len(linkcols) == 4
        assert linkcols[0] == [
            '<a href="javascript:transModal(1, mytable)">T1foo</a>',
            '<a href="/accounts/1">BankOne (1)</a>',
            '<a href="/budgets/1">Periodic1 (1)</a>',
            '<a href="/scheduled/1">Yes (1)</a>',
            '<a href="javascript:txnReconcileModal(1)">Yes (1)</a>'
        ]
        assert linkcols[1] == [
            '<a href="javascript:transModal(2, mytable)">T2</a>',
            '<a href="/accounts/2">BankTwoStale (2)</a>',
            '<a href="/budgets/4">Standing1 (4)</a>',
            '<a href="/scheduled/3">Yes (3)</a>',
            '&nbsp;'
        ]
        assert linkcols[2] == [
            '<a href="javascript:transModal(3, mytable)">T3</a>',
            '<a href="/accounts/3">CreditOne (3)</a>',
            '<a href="/budgets/2">Periodic2 (2)</a>',
            '&nbsp;',
            '&nbsp;'
        ]
        assert linkcols[3] == [
            '<a href="javascript:transModal(4, mytable)">T4split</a>',
            '<a href="/accounts/3">CreditOne (3)</a>',
            '<a href="/budgets/2">Periodic2 (2) ($222.22)</a><br>'
            '<a href="/budgets/1">Periodic1 (1) ($100.10)</a>',
            '&nbsp;',
            '&nbsp;'
        ]

    def test_acct_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        acct_filter = Select(selenium.find_element(By.ID, 'account_filter'))
        # find the options
        opts = []
        for o in acct_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]

    def test_acct_filter(self, selenium):
        p1trans = [
            'T1foo',
            'T2',
            'T3',
            'T4split'
        ]
        self.get(selenium, self.baseurl + '/transactions')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element(By.ID, 'account_filter'))
        # select BankOne (1)
        acct_filter.select_by_value('1')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T1foo']
        # select back to all
        acct_filter.select_by_value('None')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == p1trans

    def test_budg_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        budg_filter = Select(selenium.find_element(By.ID, 'budget_filter'))
        # find the options
        opts = []
        for o in budg_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]

    def test_budg_filter(self, selenium):
        p1trans = [
            'T1foo',
            'T2',
            'T3',
            'T4split'
        ]
        self.get(selenium, self.baseurl + '/transactions')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == p1trans
        budg_filter = Select(selenium.find_element(By.ID, 'budget_filter'))
        # select Periodic2 (2)
        budg_filter.select_by_value('2')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T3', 'T4split']
        # select Standing1 (4)
        budg_filter.select_by_value('4')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T2']
        # select back to all
        budg_filter.select_by_value('None')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == p1trans

    def test_search(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('foo')
        self.wait_for_jquery_done(selenium)
        # Wait for DataTables search to filter results (has debounce delay)
        self.wait_for_datatable_rows(selenium, 'table-transactions', 1)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-transactions')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == ['T1foo']


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransModalByURL(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert t.actual_amount == Decimal('222.22')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 2
        assert t.budget_transactions[0].amount == Decimal('222.22')
        assert t.sales_tax == Decimal('12.34')

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions/3')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 3'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '3'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == (
            dtnow() - timedelta(days=2)).date().strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '222.22'
        assert body.find_element(By.ID,
                                 'trans_frm_sales_tax').get_attribute('value') == '12.34'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'T3'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '3'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert selenium.find_element(By.ID,
                                     'trans_frm_notes').get_attribute('value') == 'notesT3'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransModal(AcceptanceHelper):

    def test_00_simple_modal_verify_db(self, testdb):
        t = testdb.query(Transaction).get(2)
        assert t is not None
        assert t.description == 'T2'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('-333.33')
        assert t.budgeted_amount == Decimal('-333.33')
        assert t.planned_budget_id == 4
        assert t.account_id == 2
        assert t.scheduled_trans_id == 3
        assert t.notes == 'notesT2'
        assert t.sales_tax == Decimal('0.0')
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 4
        assert t.budget_transactions[0].amount == Decimal('-333.33')
        assert testdb.query(Budget).get(4).current_balance == Decimal('1284.23')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9482.29')

    def test_01_simple_modal_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T2"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 2'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == dtnow().date(
        ).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '-333.33'
        assert body.find_element(By.ID,
                                 'trans_frm_sales_tax').get_attribute('value') == '0'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'T2'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '2'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '4'
        assert selenium.find_element(By.ID,
                                     'trans_frm_notes').get_attribute('value') == 'notesT2'

    def test_02_simple_modal_modal_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T2"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 2'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '2'
        d = body.find_element(By.ID, 'trans_frm_date')
        d.clear()
        d.send_keys(
            (dtnow() - timedelta(days=3)).date().strftime('%Y-%m-%d')
        )
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('-123.45')
        desc = body.find_element(By.ID, 'trans_frm_description')
        desc.send_keys('edited')
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        acct_sel.select_by_value('4')
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        budget_sel.select_by_value('5')
        notes = selenium.find_element(By.ID, 'trans_frm_notes')
        notes.send_keys('edited')
        tax = body.find_element(By.ID, 'trans_frm_sales_tax')
        tax.clear()
        tax.send_keys('45.67')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 2 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'T2edited' in texts

    def test_03_simple_modal_verify_db(self, testdb):
        t = testdb.query(Transaction).get(2)
        assert t is not None
        assert t.description == 'T2edited'
        assert t.date == (dtnow() - timedelta(days=3)).date()
        assert t.actual_amount == Decimal('-123.45')
        assert t.budgeted_amount == Decimal('-333.33')
        assert t.account_id == 4
        assert t.planned_budget_id == 4
        assert t.scheduled_trans_id == 3
        assert t.notes == 'notesT2edited'
        assert t.sales_tax == Decimal('45.67')
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 5
        assert t.budget_transactions[0].amount == Decimal('-123.45')
        assert testdb.query(Budget).get(4).current_balance == Decimal('950.90')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9605.74')

    def test_10_cant_edit_reconciled_verify_db(self, testdb):
        t = testdb.query(Transaction).get(1)
        assert t is not None
        assert t.description == 'T1foo'
        assert t.date == (dtnow() + timedelta(days=4)).date()
        assert t.actual_amount == Decimal('111.13')
        assert t.budgeted_amount == Decimal('111.11')
        assert t.account_id == 1
        assert t.planned_budget_id == 1
        assert t.scheduled_trans_id == 1
        assert t.notes == 'notesT1'
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 1
        assert t.budget_transactions[0].amount == Decimal('111.13')

    def test_11_cant_edit_reconciled_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T1foo"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 1'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '1'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == (
            dtnow() + timedelta(days=4)).date().strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '111.13'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'T1foo'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert selenium.find_element(By.ID,
                                     'trans_frm_notes').get_attribute('value') == 'notesT1'

    def test_12_cant_edit_reconciled_modal_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T1foo"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 1'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '1'
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-danger' in x.get_attribute('class')
        assert x.text.strip() == 'Server Error: Transaction 1 is already ' \
                                 'reconciled; cannot be edited.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_22_modal_add(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.ID, 'btn_add_trans')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        date_input = body.find_element(By.ID, 'trans_frm_date')
        # BEGIN select the 15th of this month from the popup
        dnow = dtnow()
        expected_date = date(year=dnow.year, month=dnow.month, day=15)
        date_input.click()
        date_number = body.find_element(By.XPATH,
                                        '//td[@class="day" and text()="15"]'
                                        )
        date_number.click()
        assert date_input.get_attribute(
            'value') == expected_date.strftime('%Y-%m-%d')
        # END date chooser popup
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        desc = body.find_element(By.ID, 'trans_frm_description')
        desc.send_keys('NewTrans5')
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        budget_sel.select_by_value('2')
        notes = selenium.find_element(By.ID, 'trans_frm_notes')
        notes.send_keys('NewTransNotes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 5 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans5' in texts

    def test_23_modal_add_verify_db(self, testdb):
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'NewTrans5'
        dnow = dtnow()
        assert t.date == date(year=dnow.year, month=dnow.month, day=15)
        assert t.actual_amount == Decimal('123.45')
        assert t.budgeted_amount is None
        assert t.account_id == 1
        assert t.planned_budget_id is None
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewTransNotes'
        assert t.sales_tax == Decimal('0.0')
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 2
        assert t.budget_transactions[0].amount == Decimal('123.45')

    def test_31_verify_index_budgets_table(self, base_url, selenium):
        self.get(selenium, base_url + '/')
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$950.90'],
            ['Standing2 (5)', '$9,605.74']
        ]

    def test_32_modal_add(self, base_url, selenium):
        """Test that updating a transaction against a standing budget actually
            updates the balance on the standing budget."""
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.ID, 'btn_add_trans')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        date_input = body.find_element(By.ID, 'trans_frm_date')
        assert date_input.get_attribute(
            'value') == dtnow().strftime('%Y-%m-%d')
        # END date chooser popup
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('345.67')
        desc = body.find_element(By.ID, 'trans_frm_description')
        desc.send_keys('NewTrans6')
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        budget_sel.select_by_value('5')
        notes = selenium.find_element(By.ID, 'trans_frm_notes')
        notes.send_keys('NewTransNotes')
        tax = selenium.find_element(By.ID, 'trans_frm_sales_tax')
        tax.send_keys('67.89')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 6 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans6' in texts

    def test_33_verify_db(self, testdb):
        t = testdb.query(Transaction).get(6)
        assert t is not None
        assert t.description == 'NewTrans6'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('345.67')
        assert t.budgeted_amount is None
        assert t.account_id == 1
        assert t.planned_budget_id is None
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewTransNotes'
        assert t.sales_tax == Decimal('67.89')
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 5
        assert t.budget_transactions[0].amount == Decimal('345.67')

    def test_34_verify_index_budgets_table(self, base_url, selenium):
        self.get(selenium, base_url + '/')
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$950.90'],
            ['Standing2 (5)', '$9,260.07']
        ]

    def test_40_simple_modal_verify_db(self, testdb):
        assert testdb.query(Budget).get(4).current_balance == Decimal('950.90')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9260.07')

    def test_41_modal_edit_change_between_standing(self, base_url, selenium):
        """
        test moving a transaction from one standing budget to another.
        this is as much a test of biweeklybudget.db_event_handlers as the
        transactions view.
        """
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="NewTrans6"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 6'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '6'
        amt = body.find_element(By.ID, 'trans_frm_amount')
        assert amt.get_attribute('value') == '345.67'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '5'
        budget_sel.select_by_value('4')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 6 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()

    def test_42_simple_modal_verify_db(self, testdb):
        t = testdb.query(Transaction).get(6)
        assert t is not None
        assert t.description == 'NewTrans6'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('345.67')
        assert t.budgeted_amount is None
        assert t.account_id == 1
        assert t.planned_budget_id is None
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewTransNotes'
        assert len(t.budget_transactions) == 1
        assert t.budget_transactions[0].budget_id == 4
        assert t.budget_transactions[0].amount == Decimal('345.67')
        assert testdb.query(Budget).get(4).current_balance == Decimal('605.23')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9605.74')


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransModalDoesNotShowInactiveBudgets(AcceptanceHelper):

    def test_00_add_transactions(self, testdb):
        # Transaction 5 - split with active and inactive
        testdb.add(Transaction(
            account_id=1,
            budget_amounts={
                testdb.query(Budget).get(1): Decimal('100.10'),
                testdb.query(Budget).get(2): Decimal('102.11'),
                testdb.query(Budget).get(3): Decimal('120.11')
            },
            date=dtnow().date(),
            description='InactiveBudgets1',
            notes='InactiveBudgets Txn1'
        ))
        # Transaction 6 - Active only
        testdb.add(Transaction(
            account_id=1,
            budget_amounts={
                testdb.query(Budget).get(1): Decimal('322.32')
            },
            date=dtnow().date(),
            description='InactiveBudgets2',
            notes='InactiveBudgets Txn2'
        ))
        # Transaction 7 - Inactive only
        testdb.add(Transaction(
            account_id=1,
            budget_amounts={
                testdb.query(Budget).get(3): Decimal('322.32')
            },
            date=dtnow().date(),
            description='InactiveBudgets3',
            notes='InactiveBudgets Txn3'
        ))
        testdb.commit()

    def test_01_verify_db(self, testdb):
        # Transaction 5 - split with active and inactive
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'InactiveBudgets1'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('322.32')
        assert t.account_id == 1
        assert t.notes == 'InactiveBudgets Txn1'
        assert len(t.budget_transactions) == 3
        assert {x.budget_id: x.amount for x in t.budget_transactions} == {
            1: Decimal('100.10'),
            2: Decimal('102.11'),
            3: Decimal('120.11')
        }
        # Transaction 6 - Active only
        t = testdb.query(Transaction).get(6)
        assert t is not None
        assert t.description == 'InactiveBudgets2'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('322.32')
        assert t.account_id == 1
        assert t.notes == 'InactiveBudgets Txn2'
        assert len(t.budget_transactions) == 1
        assert {x.budget_id: x.amount for x in t.budget_transactions} == {
            1: Decimal('322.32')
        }
        # Transaction 7 - Inactive only
        t = testdb.query(Transaction).get(7)
        assert t is not None
        assert t.description == 'InactiveBudgets3'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('322.32')
        assert t.account_id == 1
        assert t.notes == 'InactiveBudgets Txn3'
        assert len(t.budget_transactions) == 1
        assert {x.budget_id: x.amount for x in t.budget_transactions} == {
            3: Decimal('322.32')
        }

    def test_02_modal_populate_split(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="InactiveBudgets1"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 5'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '5'
        # assert budget split items are shown and checkbox is checked
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        # there should be three split budget input groups
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 3
        # BUDGET 0
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_0'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['3', 'Periodic3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '3'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_0').get_attribute('value') == '120.11'
        # BUDGET 1
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_1'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_1').get_attribute('value') == '102.11'
        # BUDGET 2
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_2'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_2').get_attribute('value') == '100.1'

    def test_03_modal_populate_nonsplit_active(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="InactiveBudgets2"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 6'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '6'
        # NOT Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False
        amt = body.find_element(By.ID, 'trans_frm_amount')
        assert amt.get_attribute('value') == '322.32'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]

    def test_04_modal_populate_nonsplit_inactive(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="InactiveBudgets3"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 7'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '7'
        # NOT Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False
        amt = body.find_element(By.ID, 'trans_frm_amount')
        assert amt.get_attribute('value') == '322.32'
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '3'
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['3', 'Periodic3 Inactive']
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransReconciledModal(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(TxnReconcile).get(1)
        assert t.ofx_account_id == 1
        assert t.ofx_fitid == 'BankOne.0.1'
        assert t.txn_id == 1
        assert t.rule_id is None
        assert t.note == 'reconcile notes'
        assert t.reconciled_at == datetime(2017, 4, 10, 8, 9, 11, tzinfo=UTC)

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH,
                                     '//a[@href="javascript:txnReconcileModal(1)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile 1'
        dl = body.find_element(By.TAG_NAME, 'dl')
        assert dl.get_attribute('innerHTML') == '\n' \
            '<dt>Date Reconciled</dt><dd>2017-04-10 08:09:11 UTC</dd>\n' \
            '<dt>Note</dt><dd>reconcile notes</dd>\n' \
            '<dt>Rule</dt><dd>null</dd>\n'
        trans_tbl = body.find_element(By.ID, 'txnReconcileModal-trans')
        trans_texts = self.tbody2textlist(trans_tbl)
        assert trans_texts == [
            ['Transaction'],
            [
                'Date',
                (dtnow() + timedelta(days=4)).strftime('%Y-%m-%d')
            ],
            ['Amount', '$111.13'],
            ['Budgeted Amount', '$111.11'],
            ['Description', 'T1foo'],
            ['Account', 'BankOne (1)'],
            ['Budget', 'Periodic1 (1)'],
            ['Notes', 'notesT1'],
            ['Scheduled?', 'Yes (1)']
        ]
        trans_elems = self.tbody2elemlist(trans_tbl)
        assert trans_elems[5][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
        assert trans_elems[6][1].get_attribute('innerHTML') == '<a href=' \
            '"/budgets/1">Periodic1 (1)</a>'
        assert trans_elems[8][1].get_attribute('innerHTML') == '<a href=' \
            '"/scheduled/1">Yes (1)</a>'
        ofx_tbl = body.find_element(By.ID, 'txnReconcileModal-ofx')
        ofx_texts = self.tbody2textlist(ofx_tbl)
        assert ofx_texts == [
            ['OFX Transaction'],
            ['Account', 'BankOne (1)'],
            ['FITID', 'BankOne.0.1'],
            ['Date Posted', (dtnow() - timedelta(days=6)).strftime('%Y-%m-%d')],
            ['Amount', '-$20.00'],
            ['Name', 'Late Fee'],
            ['Memo', ''],
            ['Type', 'Debit'],
            ['Description', ''],
            ['Notes', ''],
            ['Checknum', ''],
            ['MCC', ''],
            ['SIC', ''],
            ['OFX Statement'],
            ['ID', '1'],
            ['Date', (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')],
            ['Filename', '/stmt/BankOne/0'],
            [
                'File mtime',
                (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')
            ],
            ['Ledger Balance', '$12,345.67']
        ]
        ofx_elems = self.tbody2elemlist(ofx_tbl)
        assert ofx_elems[1][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'

    def test_2_split_trans(self, testdb):
        b1 = testdb.query(Budget).get(1)  # Periodic1
        b2 = testdb.query(Budget).get(2)  # Periodic2
        t = testdb.query(Transaction).get(1)
        t.set_budget_amounts({
            b1: Decimal('110.02'),
            b2: Decimal('1.11')
        })
        testdb.commit()

    def test_3_split_trans_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH,
                                     '//a[@href="javascript:txnReconcileModal(1)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile 1'
        dl = body.find_element(By.TAG_NAME, 'dl')
        assert dl.get_attribute('innerHTML') == '\n' \
            '<dt>Date Reconciled</dt><dd>2017-04-10 08:09:11 UTC</dd>\n' \
            '<dt>Note</dt><dd>reconcile notes</dd>\n' \
            '<dt>Rule</dt><dd>null</dd>\n'
        trans_tbl = body.find_element(By.ID, 'txnReconcileModal-trans')
        trans_texts = self.tbody2textlist(trans_tbl)
        assert trans_texts == [
            ['Transaction'],
            [
                'Date',
                (dtnow() + timedelta(days=4)).strftime('%Y-%m-%d')
            ],
            ['Amount', '$111.13'],
            ['Budgeted Amount', '$111.11'],
            ['Description', 'T1foo'],
            ['Account', 'BankOne (1)'],
            ['Budget', 'Periodic1 (1) ($110.02)\nPeriodic2 (2) ($1.11)'],
            ['Notes', 'notesT1'],
            ['Scheduled?', 'Yes (1)']
        ]
        trans_elems = self.tbody2elemlist(trans_tbl)
        assert trans_elems[5][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
        assert trans_elems[6][1].get_attribute('innerHTML') == '<a href=' \
            '"/budgets/1">Periodic1 (1) ($110.02)</a><br><a href=' \
            '"/budgets/2">Periodic2 (2) ($1.11)</a>'
        assert trans_elems[8][1].get_attribute('innerHTML') == '<a href=' \
            '"/scheduled/1">Yes (1)</a>'
        ofx_tbl = body.find_element(By.ID, 'txnReconcileModal-ofx')
        ofx_texts = self.tbody2textlist(ofx_tbl)
        assert ofx_texts == [
            ['OFX Transaction'],
            ['Account', 'BankOne (1)'],
            ['FITID', 'BankOne.0.1'],
            ['Date Posted', (dtnow() - timedelta(days=6)).strftime('%Y-%m-%d')],
            ['Amount', '-$20.00'],
            ['Name', 'Late Fee'],
            ['Memo', ''],
            ['Type', 'Debit'],
            ['Description', ''],
            ['Notes', ''],
            ['Checknum', ''],
            ['MCC', ''],
            ['SIC', ''],
            ['OFX Statement'],
            ['ID', '1'],
            ['Date', (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')],
            ['Filename', '/stmt/BankOne/0'],
            [
                'File mtime',
                (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')
            ],
            ['Ledger Balance', '$12,345.67']
        ]
        ofx_elems = self.tbody2elemlist(ofx_tbl)
        assert ofx_elems[1][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransModalBudgetSplits(AcceptanceHelper):

    def test_01_verify_db(self, testdb):
        t = testdb.query(Transaction).get(4)
        assert t is not None
        assert t.description == 'T4split'
        assert t.date == (dtnow() - timedelta(days=35)).date()
        assert t.actual_amount == Decimal('322.32')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT4split'
        assert t.sales_tax == Decimal('34.56')
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            1: Decimal('100.10'),
            2: Decimal('222.22')
        }

    def test_02_resave_transaction(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'id': '4',
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {
                    '2': '222.22',
                    '1': '100.10'
                },
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': True,
            'success_message': 'Successfully saved Transaction 4  in database.',
            'trans_id': 4
        }

    def test_03_verify_db(self, testdb):
        t = testdb.query(Transaction).get(4)
        assert t is not None
        assert t.description == 'T4split'
        assert t.date == (dtnow() - timedelta(days=35)).date()
        assert t.actual_amount == Decimal('322.32')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT4split'
        assert t.sales_tax == Decimal('34.56')
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            1: Decimal('100.10'),
            2: Decimal('222.22')
        }

    def test_10_backend_validation_amounts(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'id': '4',
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {
                    '2': '422.32'
                },
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': False,
            'errors': {
                'account': [],
                'amount': [],
                'budgets': [
                    'Sum of all budget amounts (422.32) must equal '
                    'Transaction amount (322.32).'
                ],
                'date': [],
                'description': [],
                'id': [],
                'notes': [],
                'sales_tax': []
            }
        }

    def test_11_backend_validation_amounts(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'id': '4',
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {
                    '1': '222.32',
                    '2': '200.12'
                },
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': False,
            'errors': {
                'account': [],
                'amount': [],
                'budgets': [
                    'Sum of all budget amounts (422.44) must equal '
                    'Transaction amount (322.32).'
                ],
                'date': [],
                'description': [],
                'id': [],
                'notes': [],
                'sales_tax': []
            }
        }

    def test_12_backend_validation_no_budgets(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'id': '4',
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {},
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': False,
            'errors': {
                'account': [],
                'amount': [],
                'budgets': [
                    'Transactions must have a budget.'
                ],
                'date': [],
                'description': [],
                'id': [],
                'notes': [],
                'sales_tax': []
            }
        }

    def test_13_backend_validation_invalid_budget_id(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {'99': '322.32'},
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': False,
            'errors': {
                'account': [],
                'amount': [],
                'budgets': [
                    'Budget ID 99 is invalid.'
                ],
                'date': [],
                'description': [],
                'notes': [],
                'sales_tax': []
            }
        }

    def test_14_backend_validation_inactive_budget(self, base_url):
        res = requests.post(
            base_url + '/forms/transaction',
            json={
                'date': (dtnow() - timedelta(days=35)).strftime('%Y-%m-%d'),
                'amount': '322.32',
                'description': 'T4split',
                'notes': 'notesT4split',
                'account': '3',
                'budgets': {'3': '322.32'},
                'sales_tax': '34.56'
            }
        )
        assert res.status_code == 200
        assert res.json() == {
            'success': False,
            'errors': {
                'account': [],
                'amount': [],
                'budgets': [
                    'New transactions cannot use an inactive budget '
                    '(Periodic3 Inactive).'
                ],
                'date': [],
                'description': [],
                'notes': [],
                'sales_tax': []
            }
        }

    def validation_count_increased(self, driver, previous):
        c = driver.execute_script('return validation_count;')
        return c > previous

    def assert_budget_split_has_error(self, driver, msg):
        # get validate count
        c = driver.execute_script('return validation_count;')
        # change focus
        driver.find_element(By.ID, 'trans_frm_description').click()
        # wait for validate count to increase
        try:
            WebDriverWait(driver, 5).until(
                lambda x: self.validation_count_increased(driver, c)
            )
        except TimeoutException:
            pass
        assert driver.find_element(By.ID, 'budget-split-feedback').text == msg
        assert driver.find_element(By.ID,
                                   'modalSaveButton').is_enabled() is False

    def assert_budget_split_does_not_have_error(self, driver):
        # get validate count
        c = driver.execute_script('return validation_count;')
        # change focus
        driver.find_element(By.ID, 'trans_frm_description').click()
        # wait for validate count to increase
        try:
            WebDriverWait(driver, 5).until(
                lambda x: self.validation_count_increased(driver, c)
            )
        except TimeoutException:
            pass
        assert driver.find_element(By.ID, 'budget-split-feedback').text == ''
        assert driver.find_element(By.ID, 'modalSaveButton').is_enabled()

    def test_20_modal_frontend_validation(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.ID, 'btn_add_trans')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        # set an amount
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('200.22')
        # assert budget split items are hidden and checkbox is unchecked
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False
        # check the budget split checkbox
        selenium.find_element(By.ID, 'trans_frm_is_split').click()
        # assert budget split items are shown and checkbox is checked
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        # there should be two split budget input groups
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 2
        self.assert_budget_split_does_not_have_error(selenium)
        # Select 2 different budgets and valid amounts
        Select(
            body.find_element(By.ID, 'trans_frm_budget_0')).select_by_value('1')
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_0')
        tmp.clear()
        tmp.send_keys('100')
        Select(
            body.find_element(By.ID, 'trans_frm_budget_1')).select_by_value('2')
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_1')
        tmp.clear()
        tmp.send_keys('100.22')
        self.assert_budget_split_does_not_have_error(selenium)
        # change one amount
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_1')
        tmp.clear()
        tmp.send_keys('100.00')
        self.assert_budget_split_has_error(
            selenium,
            'Error: Sum of budget allocations (200.0000) must equal '
            'transaction amount (200.2200).'
        )
        # fix the amount
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_1')
        tmp.clear()
        tmp.send_keys('100.22')
        self.assert_budget_split_does_not_have_error(selenium)
        # change one budget to the same as the other
        Select(
            body.find_element(By.ID, 'trans_frm_budget_1')).select_by_value('1')
        self.assert_budget_split_has_error(
            selenium,
            'Error: A given budget may only be specified once.'
        )
        # fix the budget
        Select(
            body.find_element(By.ID, 'trans_frm_budget_1')).select_by_value('2')
        self.assert_budget_split_does_not_have_error(selenium)
        # click "Add Budget" link
        self.try_click(
            selenium, selenium.find_element(By.ID, 'trans_frm_add_budget_link')
        )
        # there should be three split budget input groups
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 3
        # decrease an amount in one of the previous groups
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_1')
        tmp.clear()
        tmp.send_keys('50.11')
        self.assert_budget_split_has_error(
            selenium,
            'Error: Sum of budget allocations (150.1100) must equal '
            'transaction amount (200.2200).'
        )
        # add difference to amount in the third budget group
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_2')
        tmp.clear()
        tmp.send_keys('50.11')
        self.assert_budget_split_does_not_have_error(selenium)
        # select budget in third group, same as second
        Select(
            body.find_element(By.ID, 'trans_frm_budget_2')).select_by_value('2')
        self.assert_budget_split_has_error(
            selenium,
            'Error: A given budget may only be specified once.'
        )
        # change budget in third group to a unique one
        Select(
            body.find_element(By.ID, 'trans_frm_budget_2')).select_by_value('4')
        self.assert_budget_split_does_not_have_error(selenium)
        # uncheck the Budget Split checkbox
        selenium.find_element(By.ID, 'trans_frm_is_split').click()
        # assert budget split items are hidden and checkbox is unchecked
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed()
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False

    def test_30_verify_db_before(self, testdb):
        t = testdb.query(Transaction).get(4)
        assert t is not None
        assert t.description == 'T4split'
        assert t.date == (dtnow() - timedelta(days=35)).date()
        assert t.actual_amount == Decimal('322.32')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT4split'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            1: Decimal('100.10'),
            2: Decimal('222.22')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 4

    def test_31_split_2_modal_populate(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T4split"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 4'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '4'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == (
                dtnow() - timedelta(days=35)
            ).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '322.32'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'T4split'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '3'
        # Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 2
        # BUDGET 0
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_0'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_0').get_attribute('value') == '222.22'
        # BUDGET 1
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_1'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_1').get_attribute('value') == '100.1'
        assert selenium.find_element(By.ID,
                                     'trans_frm_notes').get_attribute('value') == 'notesT4split'

    def test_32_new_split_trans(self, base_url, selenium, testdb):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.ID, 'btn_add_trans')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        date_input = body.find_element(By.ID, 'trans_frm_date')
        assert date_input.get_attribute(
            'value') == dtnow().strftime('%Y-%m-%d')
        # END date chooser popup
        amt = body.find_element(By.ID, 'trans_frm_amount')
        amt.clear()
        amt.send_keys('375.00')
        desc = body.find_element(By.ID, 'trans_frm_description')
        desc.send_keys('NewTrans5')
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        # check the budget split checkbox
        selenium.find_element(By.ID, 'trans_frm_is_split').click()
        # assert budget split items are shown and checkbox is checked
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        # there should be two split budget input groups
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 2
        # set the budgets and amounts
        Select(
            body.find_element(By.ID, 'trans_frm_budget_0')).select_by_value('1')
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_0')
        tmp.clear()
        tmp.send_keys('100.00')
        Select(
            body.find_element(By.ID, 'trans_frm_budget_1')).select_by_value('2')
        # the next value should be populated automatically
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_1').get_attribute('value') == '275.00'
        tmp = body.find_element(By.ID, 'trans_frm_budget_amount_1')
        tmp.clear()
        tmp.send_keys('200')
        # change focus
        body.find_element(By.ID, 'trans_frm_budget_amount_0').send_keys('')
        # add a row
        self.try_click(
            selenium, selenium.find_element(By.ID, 'trans_frm_add_budget_link')
        )
        # there should be three split budget input groups
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 3
        # the amount should be populated automatically
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_2').get_attribute('value') == '75.00'
        # fill in the third row
        Select(
            body.find_element(By.ID, 'trans_frm_budget_2')).select_by_value('4')
        self.assert_budget_split_does_not_have_error(selenium)
        notes = selenium.find_element(By.ID, 'trans_frm_notes')
        notes.send_keys('NewSplitTransNotes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 5 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans5' in texts
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'NewTrans5'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('375')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 1
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewSplitTransNotes'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            1: Decimal('100'),
            2: Decimal('200'),
            4: Decimal('75')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5

    def test_33_change_split_trans(self, base_url, selenium, testdb):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="NewTrans5"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 5'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '5'
        assert body.find_element(By.ID,
                                 'trans_frm_date'
                                 ).get_attribute('value') == dtnow().strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '375'
        assert body.find_element(By.ID,
                                 'trans_frm_description').get_attribute('value') == 'NewTrans5'
        acct_sel = Select(body.find_element(By.ID, 'trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        # Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 3
        # BUDGET 0
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_0'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_0').get_attribute('value') == '200'
        # BUDGET 1
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_1'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_1').get_attribute('value') == '100'
        # BUDGET 2
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget_2'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '4'
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_2').get_attribute('value') == '75'
        elem = selenium.find_element(By.ID, 'trans_frm_notes')
        assert elem.get_attribute('value') == 'NewSplitTransNotes'
        # Ok, now edit it...
        Select(body.find_element(By.ID,
                                 'trans_frm_budget_1')).select_by_value('None')
        body.find_element(By.ID, 'trans_frm_budget_amount_1').clear()
        budget_amt = body.find_element(By.ID, 'trans_frm_budget_amount_0')
        budget_amt.clear()
        budget_amt.send_keys('300')
        self.assert_budget_split_does_not_have_error(selenium)
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 5 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element(By.ID, 'table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans5' in texts
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'NewTrans5'
        assert t.date == dtnow().date()
        assert t.actual_amount == Decimal('375')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 1
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewSplitTransNotes'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            2: Decimal('300'),
            4: Decimal('75')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5

    def test_34_existing_trans_to_split(self, base_url, selenium, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert t.actual_amount == Decimal('222.22')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            2: Decimal('222.22')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T3"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 3'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '3'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == (
                dtnow() - timedelta(days=2)
            ).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '222.22'
        # NOT Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False
        # Ok, click to split it...
        self.try_click(
            selenium, selenium.find_element(By.ID, 'trans_frm_is_split')
        )
        # Should be split now...
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected()
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 2
        # Verify that initial budget was set
        assert Select(
            body.find_element(By.ID, 'trans_frm_budget_0')
        ).first_selected_option.get_attribute('value') == '2'
        # Verify that amount has been set
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_0').get_attribute('value') == '222.22'
        # Set the amount
        budget_amt = body.find_element(By.ID, 'trans_frm_budget_amount_0')
        budget_amt.clear()
        budget_amt.send_keys('100.02')
        # select the second budget
        Select(body.find_element(By.ID,
                                 'trans_frm_budget_1')).select_by_value('4')
        # Verify that second amount is set
        assert body.find_element(By.ID,
                                 'trans_frm_budget_amount_1').get_attribute('value') == '122.20'
        self.assert_budget_split_does_not_have_error(selenium)
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 3 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_35_verify_db(self, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert t.actual_amount == Decimal('222.22')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            2: Decimal('100.02'),
            4: Decimal('122.20')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5

    def test_36_existing_split_trans_to_not(self, base_url, selenium, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert t.actual_amount == Decimal('222.22')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            2: Decimal('100.02'),
            4: Decimal('122.20')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH, '//a[text()="T3"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 3'
        assert body.find_element(By.ID,
                                 'trans_frm_id').get_attribute('value') == '3'
        assert body.find_element(By.ID,
                                 'trans_frm_date').get_attribute('value') == (
                       dtnow() - timedelta(days=2)
               ).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'trans_frm_amount').get_attribute('value') == '222.22'
        # Should be split...
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected()
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed()
        assert len(
            selenium.find_elements(By.CLASS_NAME, 'budget_split_row')
        ) == 2
        # Ok, click to un-split it...
        self.try_click(
            selenium, selenium.find_element(By.ID, 'trans_frm_is_split')
        )
        # NOT Split Budget
        assert selenium.find_element(By.ID,
                                     'trans_frm_is_split').is_selected() is False
        assert selenium.find_element(By.ID,
                                     'trans_frm_budget_group').is_displayed() is True
        assert selenium.find_element(By.ID,
                                     'trans_frm_split_budget_container').is_displayed() is False
        # Ok, now edit it...
        budget_sel = Select(body.find_element(By.ID, 'trans_frm_budget'))
        budget_sel.select_by_value('2')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 3 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_37_verify_db(self, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert t.actual_amount == Decimal('222.22')
        assert t.budgeted_amount is None
        assert t.planned_budget_id is None
        assert t.account_id == 3
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'
        assert {bt.budget_id: bt.amount for bt in t.budget_transactions} == {
            2: Decimal('222.22')
        }
        assert max([
            tx.id for tx in testdb.query(Transaction).all()
        ]) == 5



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import Column, Integer, Numeric, Boolean, String
from biweeklybudget.models.base import Base, ModelAsDict


class Budget(Base, ModelAsDict):

    __tablename__ = 'budgets'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Whether the budget is standing (long-running) or periodic (resets each
    #: pay period or budget cycle)
    is_periodic = Column(Boolean, default=True)

    #: name of the budget
    name = Column(String(50), unique=True, index=True)

    #: description
    description = Column(String(254))

    #: starting balance for periodic budgets
    starting_balance = Column(Numeric(precision=10, scale=4))

    #: current balance for standing budgets
    current_balance = Column(Numeric(precision=10, scale=4))

    #: whether active or historical
    is_active = Column(Boolean, default=True)

    #: whether this is an Income budget (True) or expense (False).
    is_income = Column(Boolean, default=False)

    #: whether or not to omit this budget from spending graphs
    omit_from_graphs = Column(Boolean, default=False)

    def __repr__(self):
        return "<Budget(id=%s, name=%s)>" % (
            self.id, self.name
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import importlib
import logging
from datetime import timedelta, datetime
from babel.numbers import validate_currency, UnknownCurrencyError
from babel import Locale, UnknownLocaleError

logger = logging.getLogger(__name__)

_REQUIRED_VARS = [
    'DB_CONNSTRING',
    'DEFAULT_ACCOUNT_ID',
    'PAY_PERIOD_START_DATE',
    'RECONCILE_BEGIN_DATE',
    'STALE_DATA_TIMEDELTA',
    'CURRENCY_CODE'
]

_DATE_VARS = [
    'PAY_PERIOD_START_DATE',
    'RECONCILE_BEGIN_DATE'
]
_TIMEDELTA_VARS = [
    'STALE_DATA_TIMEDELTA'
]
_INT_VARS = [
    'DEFAULT_ACCOUNT_ID',
    'FUEL_BUDGET_ID',
    'BIWEEKLYBUDGET_TEST_TIMESTAMP'
]
_STRING_VARS = [
    'DB_CONNSTRING',
    'STATEMENTS_SAVE_PATH',
    'TOKEN_PATH',
    'VAULT_ADDR',
    'LOCALE_NAME',
    'CURRENCY_CODE',
    'FUEL_VOLUME_UNIT',
    'FUEL_VOLUME_ABBREVIATION',
    'FUEL_ECO_ABBREVIATION',
    'PLAID_CLIENT_ID',
    'PLAID_SECRET',
    'PLAID_ENV',
    'PLAID_PRODUCTS',
    'PLAID_COUNTRY_CODES',
    'PLAID_USER_ID',
]

#: A `RFC 5646 / BCP 47 <https://tools.ietf.org/html/bcp47>`_ Language Tag
#: with a Region suffix to use for number (currency) formatting, i.e. "en_US",
#: "en_GB", "de_DE", etc. If this is not specified (None), it will be looked up
#: from environment variables in the following order: LC_ALL, LC_MONETARY, LANG.
#: If none of those variables are set to a valid locale name (not including
#: the "C" locale, which does not specify currency formatting) and this variable
#: is not set, the application will default to "en_US". This setting only
#: effects how monetary values are displayed in the UI, logs, etc. For further
#: information, see
#: :ref:`Currency Formatting and Localization <app_usage.l10n>`.
LOCALE_NAME = None

#: An `ISO 4217 <https://en.wikipedia.org/wiki/ISO_4217>`_ Currency Code
#: specifying the currency to use for all monetary amounts, i.e. "USD", "EUR",
#: etc. This setting only effects how monetary values are displayed in the UI,
#: logs, etc. Currently defaults to "USD". For further information, see
#: :ref:`Currency Formatting and Localization <app_usage.l10n>`.
CURRENCY_CODE = 'USD'

#: The full written name of your unit of measure for volume of fuel, to be used
#: for the Fuel Log feature. As an example, ``Gallons`` or ``Litres``.
FUEL_VOLUME_UNIT = 'Gallons'

#: Abbreviation of :py:attr:`biweeklybudget.settings.FUEL_VOLUME_UNIT`, such as
#: ``Gal.`` or ``L``.
FUEL_VOLUME_ABBREVIATION = 'Gal.'

#: The full written name of your unit of distance for fuel economy calculations
#: and the Fuel Log. As an example, ``Miles`` or ``Kilometers``.
DISTANCE_UNIT = 'Miles'

#: Abbreviation of :py:attr:`biweeklybudget.settings.DISTANCE_UNIT`, such as
#: ``Mi.`` or ``KM``.
DISTANCE_UNIT_ABBREVIATION = 'Mi.'

#: Abbreviation for your distance-per-volume fuel economy measurement,
#: such as ``MPG`` or ``KM/L``.
FUEL_ECO_ABBREVIATION = 'MPG'

#: string - SQLAlchemy database connection string. See the
#: :ref:`SQLAlchemy Database URLS docs <sqlalchemy:database_urls>`
#: for further information.
DB_CONNSTRING = None

#: int - Account ID to show first in dropdown lists. This must be the database
#: ID of a valid account.
DEFAULT_ACCOUNT_ID = 1

#: int - Budget ID to select as default when inputting Fuel Log entries. This
#: must be the database ID of a valid budget.
FUEL_BUDGET_ID = 1

#: :py:class:`datetime.date` - The starting date of one pay period (generally
#: the first pay period represented in data in this app). The dates of all pay
#: periods will be determined based on an interval from this date. This must
#: be specified in Y-m-d format (i.e. parsable by
#: :py:meth:`datetime.datetime.strptime` with ``%Y-%m-%d`` format).
PAY_PERIOD_START_DATE = None

#: :py:class:`datetime.date` - When listing unreconciled transactions that need
#: to be reconciled, any transaction before this date will be ignored. This must
#: be specified in Y-m-d format (i.e. parsable by
#: :py:meth:`datetime.datetime.strptime` with ``%Y-%m-%d`` format).
RECONCILE_BEGIN_DATE = None

#: :py:class:`datetime.timedelta` - Time interval beyond which OFX data for
#: accounts will be considered old/stale. This must be specified as a number
#: (integer) that will be converted to a number of days.
STALE_DATA_TIMEDELTA = timedelta(days=2)

#: string - *(optional)* Filesystem path to download OFX statements to, and for
#: backfill_ofx to read them from.
STATEMENTS_SAVE_PATH = None

#: string - *(optional)* Filesystem path to read Vault token from, for OFX
#: credentials.
TOKEN_PATH = None

#: string - *(optional)* Address to connect to Vault at, for OFX credentials.
VAULT_ADDR = None

#: int - FOR ACCEPTANCE TESTS ONLY - This is used to "fudge" the current time
#: to the specified integer timestamp. Used for acceptance tests only. Do NOT
#: set this outside of acceptance testing.
BIWEEKLYBUDGET_TEST_TIMESTAMP = None

#: Plaid Client ID
PLAID_CLIENT_ID = None

#: Plaid Secret (client secret)
PLAID_SECRET = None

#: Plaid environment name. Use 'sandbox' to test with Plaid's Sandbox
#: environment (username: user_good, password: pass_good). Use `development` to
#: test with live users and credentials and `production` to go live
PLAID_ENV = None

#: PLAID_PRODUCTS is a comma-separated list of products to use when initializing
#: Link. Note that this list must contain 'assets' in order for the app to be
#: able to create and retrieve asset reports.
PLAID_PRODUCTS = None

#: PLAID_COUNTRY_CODES is a comma-separated list of countries for which users
#: will be able to select institutions from.
PLAID_COUNTRY_CODES = None

#: PLAID_USER_ID is a unique per-user ID for users of Plaid applications.
#: Since this is a single-user app, we just hard-code to "1"
PLAID_USER_ID = '1'

if 'SETTINGS_MODULE' in os.environ:
    logger.debug('Attempting to import settings module %s',
                 os.environ['SETTINGS_MODULE'])
    modname = os.environ.get('SETTINGS_MODULE')
    m = importlib.import_module(modname)
    module_dict = m.__dict__
    try:
        to_import = m.__all__
    except AttributeError:
        to_import = [name for name in module_dict if not name.startswith('_')]
    v = {name: module_dict[name] for name in to_import}
    logger.debug('Import from SETTINGS_MODULE: %s', v)
    globals().update(v)
else:
    logger.debug('SETTINGS_MODULE not defined')

for varname in _STRING_VARS:
    if varname not in os.environ:
        continue
    logger.debug('Setting %S from env var: %s', varname, os.environ[varname])
    globals()[varname] = os.environ[varname]

for varname in _INT_VARS:
    if varname not in os.environ:
        continue
    try:
        value = int(os.environ[varname])
        assert os.environ[varname] == '%s' % value
    except Exception:
        raise SystemExit('ERROR: env var %s cannot parse as int' % varname)
    logger.debug('Setting %S from env var: %s', varname, value)
    globals()[varname] = value

for varname in _TIMEDELTA_VARS:
    if varname not in os.environ:
        continue
    try:
        value = int(os.environ[varname])
        assert os.environ[varname] == '%s' % value
    except Exception:
        raise SystemExit('ERROR: env var %s cannot parse as int' % varname)
    value = timedelta(days=value)
    logger.debug('Setting %S from env var: %s', varname, value)
    globals()[varname] = value

for varname in _DATE_VARS:
    if varname not in os.environ:
        continue
    try:
        value = datetime.strptime(os.environ[varname], '%Y-%m-%d')
    except Exception:
        raise SystemExit('ERROR: env var %s cannot parse as %Y-%m-%d' % varname)
    logger.debug('Setting %S from env var: %s', varname, value)
    globals()[varname] = value

for varname in _REQUIRED_VARS:
    if globals().get(varname, None) is None:
        raise SystemExit(
            'ERROR: setting or environment variable "%s" must be set' % varname
        )

# Handle the "LOCALE_NAME" variable special logic for default if not specified.
if LOCALE_NAME is None or LOCALE_NAME == 'C' or LOCALE_NAME.startswith('C.'):
    logger.debug('LOCALE_NAME unset or C locale')
    for varname in ['LC_ALL', 'LC_MONETARY', 'LANG']:
        val = os.environ.get(varname, '').strip()
        if val != '' and val != 'C' and not val.startswith('C.'):
            logger.debug(
                'Setting LOCALE_NAME to %s from env var %s', val, varname
            )
            LOCALE_NAME = val
            break
    if LOCALE_NAME is None:
        logger.debug('LOCALE_NAME not set; defaulting to en_US')
        LOCALE_NAME = 'en_US'

# Check for a valid locale
try:
    Locale.parse(LOCALE_NAME)
except UnknownLocaleError:
    raise SystemExit(
        'ERROR: LOCALE_NAME setting of "%s" is not a valid BCP 47 Language Tag.'
        ' See <https://tools.ietf.org/html/bcp47> for more information.' %
        LOCALE_NAME
    )

# Check for a valid currency code
try:
    validate_currency(CURRENCY_CODE)
except UnknownCurrencyError:
    raise SystemExit(
        'ERROR: CURRENCY_CODE setting of "%s" is not a valid currency code. See'
        ' <https://en.wikipedia.org/wiki/ISO_4217> for more information and '
        'a list of valid codes.' % CURRENCY_CODE
    )

logger.debug('Done loading settings.')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from decimal import Decimal

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestTransStandingBudgetBalanceUpdate(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        """initial state verification"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 4
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert standing.current_balance == Decimal('9482.29')
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_1_add_trans_periodic_budget(self, testdb):
        """add a transaction against a periodic budget"""
        t = Transaction(
            budget_amounts={testdb.query(Budget).get(2): Decimal('222.22')},
            budgeted_amount=Decimal('123.45'),
            description='T5',
            notes='notesT5',
            account=testdb.query(Account).get(1),
            planned_budget=testdb.query(Budget).get(2)
        )
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_2_verify_db(self, testdb):
        """verify no budget objects changed"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 5
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert standing.current_balance == Decimal('9482.29')
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_3_add_trans_standing_budget(self, testdb):
        """add a transaction against a standing budget"""
        t = Transaction(
            budget_amounts={testdb.query(Budget).get(5): Decimal('222.22')},
            budgeted_amount=Decimal('123.45'),
            description='T6',
            notes='notesT6',
            account=testdb.query(Account).get(1),
            planned_budget=testdb.query(Budget).get(5)
        )
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_4_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 6
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert standing.current_balance == Decimal('9260.07')
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_5_edit_trans_standing_budget(self, testdb):
        """edit a transaction against a standing budget"""
        t = testdb.query(Transaction).get(6)
        budg = testdb.query(Budget).get(5)
        t.set_budget_amounts({budg: Decimal('111.11')})
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_6_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 6
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert standing.current_balance == Decimal('9371.18')
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_7_edit_trans_standing_budget(self, testdb):
        """edit a transaction against a standing budget"""
        t = testdb.query(Transaction).get(6)
        t.set_budget_amounts({testdb.query(Budget).get(5): Decimal('-111.11')})
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_8_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 6
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert standing.current_balance == Decimal('9593.40')
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
class TestIsFieldsSet(AcceptanceHelper):

    def test_0_is_fields_set_by_ofxtxn_event_handler(self, testdb):
        """
        Test for
        :py:func:`~.db_event_handlers.handle_ofx_transaction_new_or_change`
        """
        acct = testdb.query(Account).get(1)
        acct.re_interest_paid = None
        testdb.commit()
        assert acct.re_interest_charge == '^interest-charge'
        assert acct.re_interest_paid is None
        assert acct.re_payment == '^(payment|thank you)'
        assert acct.re_late_fee == '^Late Fee'
        assert acct.re_other_fee == '^re-other-fee'
        stmt = testdb.query(OFXStatement).get(1)
        assert stmt.account_id == 1
        assert stmt.filename == '/stmt/BankOne/0'
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-1',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='BankOne-9-1'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is False
        assert txn.is_late_fee is False
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is False
        assert txn.is_interest_payment is False
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-2',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='re-other-fee BankOne-9-2'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is False
        assert txn.is_late_fee is False
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is True
        assert txn.is_interest_payment is False
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-3',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='Late Fee BankOne-9-3'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is False
        assert txn.is_late_fee is True
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is False
        assert txn.is_interest_payment is False
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-4',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='payment BankOne-9-4'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is True
        assert txn.is_late_fee is False
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is False
        assert txn.is_interest_payment is False
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-5',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='Thank You BankOne-9-5'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is True
        assert txn.is_late_fee is False
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is False
        assert txn.is_interest_payment is False
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='BankOne-9-6',
            trans_type='Credit',
            date_posted=stmt.ledger_bal_as_of,
            amount=Decimal('1234.56'),
            name='interest-paid'
        )
        testdb.add(txn)
        testdb.commit()
        assert txn.is_payment is False
        assert txn.is_late_fee is False
        assert txn.is_interest_charge is False
        assert txn.is_other_fee is False
        assert txn.is_interest_payment is False

    def test_1_account_re_change_triggers_update_is_fields(self, testdb):
        """
        Test for
        :py:func:`~.db_event_handlers.handle_account_re_change`
        """
        acct = testdb.query(Account).get(1)
        acct.re_interest_paid = None
        testdb.commit()
        assert acct.re_interest_charge == '^interest-charge'
        assert acct.re_interest_paid is None
        assert acct.re_payment == '^(payment|thank you)'
        assert acct.re_late_fee == '^Late Fee'
        assert acct.re_other_fee == '^re-other-fee'
        stmt = testdb.query(OFXStatement).get(1)
        assert stmt.account_id == 1
        assert stmt.filename == '/stmt/BankOne/0'
        txn1 = testdb.query(OFXTransaction).get((1, 'BankOne-9-1'))
        assert txn1.name == 'BankOne-9-1'
        assert txn1.is_payment is False
        assert txn1.is_late_fee is False
        assert txn1.is_interest_charge is False
        assert txn1.is_other_fee is False
        assert txn1.is_interest_payment is False
        txn2 = testdb.query(OFXTransaction).get((1, 'BankOne-9-2'))
        assert txn2.name == 're-other-fee BankOne-9-2'
        assert txn2.is_payment is False
        assert txn2.is_late_fee is False
        assert txn2.is_interest_charge is False
        assert txn2.is_other_fee is True
        assert txn2.is_interest_payment is False
        txn3 = testdb.query(OFXTransaction).get((1, 'BankOne-9-3'))
        assert txn3.name == 'Late Fee BankOne-9-3'
        assert txn3.is_payment is False
        assert txn3.is_late_fee is True
        assert txn3.is_interest_charge is False
        assert txn3.is_other_fee is False
        assert txn3.is_interest_payment is False
        # change the account
        acct.re_interest_charge = '^BankOne-9-1$'
        acct.re_payment = '^Late Fee'
        acct.re_late_fee = '^foobarbaz'
        testdb.commit()
        # re-confirm
        txn1 = testdb.query(OFXTransaction).get((1, 'BankOne-9-1'))
        assert txn1.name == 'BankOne-9-1'
        assert txn1.is_payment is False
        assert txn1.is_late_fee is False
        assert txn1.is_interest_charge is True
        assert txn1.is_other_fee is False
        assert txn1.is_interest_payment is False
        txn2 = testdb.query(OFXTransaction).get((1, 'BankOne-9-2'))
        assert txn2.name == 're-other-fee BankOne-9-2'
        assert txn2.is_payment is False
        assert txn2.is_late_fee is False
        assert txn2.is_interest_charge is False
        assert txn2.is_other_fee is True
        assert txn2.is_interest_payment is False
        txn3 = testdb.query(OFXTransaction).get((1, 'BankOne-9-3'))
        assert txn3.name == 'Late Fee BankOne-9-3'
        assert txn3.is_payment is True
        assert txn3.is_late_fee is False
        assert txn3.is_interest_charge is False
        assert txn3.is_other_fee is False
        assert txn3.is_interest_payment is False



from __future__ import absolute_import
from __future__ import unicode_literals
try:
    # python 3
    from io import StringIO
except ImportError:
    # python 2
    from StringIO import StringIO

from biweeklybudget.vendored.ofxclient.client import Client


def combined_download(accounts, days=60):
    """Download OFX files and combine them into one

    It expects an 'accounts' list of ofxclient.Account objects
    as well as an optional 'days' specifier which defaults to 60
    """

    client = Client(institution=None)

    out_file = StringIO()
    out_file.write(client.header())
    out_file.write('<OFX>')
    for a in accounts:
        ofx = a.download(days=days).read()
        stripped = ofx.partition('<OFX>')[2].partition('</OFX>')[0]
        out_file.write(stripped)

    out_file.write("</OFX>")
    out_file.seek(0)

    return out_file



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import re
import json
from decimal import Decimal
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.dbsetting import DBSetting
from biweeklybudget.models.account import Account, NoInterestChargedError
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.utils import dtnow
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By


def get_payoff_forms(selenium):
    form_ids = [
        s.get_attribute('id')
        for s in selenium.find_elements(By.TAG_NAME, 'form')
    ]
    res = {}
    for f_type in ['increase', 'onetime']:
        vals = {}
        id_re = re.compile(r'^payoff_' + f_type + r'_frm_(\d+)$')
        for frmid in form_ids:
            m = id_re.match(frmid)
            if m is None:
                continue
            _id = int(m.group(1))

            vals[_id] = {
                'enabled': selenium.find_element(By.ID,
                                                 'payoff_%s_frm_%s_enable' % (f_type, _id)
                                                 ).is_selected(),
                'date': selenium.find_element(By.ID,
                                              'payoff_%s_frm_%s_date' % (f_type, _id)
                                              ).get_attribute('value'),
                'amount': selenium.find_element(By.ID,
                                                'payoff_%s_frm_%s_amt' % (f_type, _id)
                                                ).get_attribute('value')
            }
        res[f_type] = vals
    return res


def set_payoff_form(selenium, f_type, idx, enabled, date_s, amt):
    is_enabled = selenium.find_element(By.ID,
                                       'payoff_%s_frm_%s_enable' % (f_type, idx)).is_selected()
    if is_enabled != enabled:
        selenium.find_element(By.ID,
                              'payoff_%s_frm_%s_enable' % (f_type, idx)).click()
        elem_id = 'payoff_%s_frm_%s_enable' % (f_type, idx)
        assert selenium.find_element(By.ID, elem_id).is_selected() == enabled
    selenium.find_element(By.ID,
                          'payoff_%s_frm_%s_date' % (f_type, idx)).clear()
    selenium.find_element(By.ID,
                          'payoff_%s_frm_%s_date' % (f_type, idx)).send_keys(date_s)
    selenium.find_element(By.ID,
                          'payoff_%s_frm_%s_amt' % (f_type, idx)).clear()
    selenium.find_element(By.ID,
                          'payoff_%s_frm_%s_amt' % (f_type, idx)).send_keys(amt)


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestCreditPayoffsNoInterest(AcceptanceHelper):

    def test_01_heading(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'ERROR - Credit Card Payoffs - BiweeklyBudget'

    def test_02_nav_menu(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_03_notifications(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'

    def test_04_danger_message(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')
        div = selenium.find_element(By.ID, 'account-interest-error-message')
        assert div is not None
        assert 'alert-danger' in div.get_attribute('class')
        text = div.get_attribute('innerHTML')
        assert 'Account Interest Error' in text
        assert 'in which case you must <a href="#" onclick="creditPayoffError' \
               'Modal(4)">manually input the interest charge ' \
               'from your last statement</a>' in text

    def test_05_verify_db(self, testdb):
        acct = testdb.query(Account).get(4)
        with pytest.raises(NoInterestChargedError):
            assert isinstance(acct.last_interest_charge, Decimal)

    def test_06_add_interest(self, selenium, base_url):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts/credit-payoff')
        link = selenium.find_element(By.XPATH,
                                     '//a[text()="manually input the interest charge from your '
                                     'last statement"]'
                                     )
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add Manual Interest Charge for Account 4'
        frm_id = selenium.find_element(By.ID, 'payoff_acct_frm_id')
        frm_filename = selenium.find_element(By.ID,
                                             'payoff_acct_frm_statement_filename'
                                             )
        frm_int_amt = selenium.find_element(By.ID,
                                            'payoff_acct_frm_interest_amt'
                                            )
        assert frm_id.get_attribute('value') == '4'
        fname = Select(frm_filename)
        # find the options
        opts = []
        for o in fname.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['/stmt/CreditTwo/0', '2017-07-26 (-$5,498.65)']
        ]
        fname.select_by_value('/stmt/CreditTwo/0')
        frm_int_amt.clear()
        frm_int_amt.send_keys('123.45')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved OFXTransaction with ' \
                                 'FITID 20170728062444-MANUAL-CCPAYOFF' \
                                 ' in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Credit Card Payoffs - BiweeklyBudget'

    def test_07_verify_db(self, testdb):
        acct = testdb.query(Account).get(4)
        assert acct.last_interest_charge == Decimal('-123.45')
        stmt = testdb.query(OFXStatement).get(7)
        ofxtxn = testdb.query(OFXTransaction).get(
            (4, '20170728062444-MANUAL-CCPAYOFF')
        )
        assert ofxtxn.account == acct
        assert ofxtxn.statement == stmt
        assert ofxtxn.fitid == '20170728062444-MANUAL-CCPAYOFF'
        assert ofxtxn.trans_type == 'debit'
        assert ofxtxn.date_posted == stmt.as_of
        assert ofxtxn.amount == Decimal('123.45')
        assert ofxtxn.name == 'Interest Charged - MANUALLY ENTERED'
        assert ofxtxn.memo is None
        assert ofxtxn.sic is None
        assert ofxtxn.mcc is None
        assert ofxtxn.checknum is None
        assert ofxtxn.description is None
        assert ofxtxn.notes is None
        assert ofxtxn.is_interest_charge is True


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'testflask')
@pytest.mark.incremental
class TestNoSettings(AcceptanceHelper):

    def test_000_setup_interest_charge_in_db(self, testdb):
        acct = testdb.query(Account).get(4)
        stmt = testdb.query(OFXStatement).get(7)
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='%s-MANUAL-CCPAYOFF' % dtnow().strftime('%Y%m%d%H%M%S'),
            trans_type='debit',
            date_posted=stmt.as_of,
            amount=Decimal('-46.9061'),
            name='Interest Charged - MANUALLY ENTERED',
            is_interest_charge=True
        )
        testdb.add(txn)
        testdb.commit()

    def test_00_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is None

    def test_01_no_config(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        assert selenium.find_element(By.ID,
                                     'payoff_frm_min_pymt').get_attribute('value') == '144.98'
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements(By.TAG_NAME, 'form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_onetime_frm_1'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            },
            'onetime': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            }
        }

    def test_01_highest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_HighestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.6 years', '$6,956.35', '$1,457.70'
            ],
            ['Totals', '$144.97', '4.6 years', '$7,919.34', '$1,468.63']
        ]

    def test_02_highest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID,
                                      'table_HighestInterestRateFirstMethod'
                                      )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.6 years', '$6,956.35', '$1,457.70'
            ],
            ['Totals', '$144.97', '4.6 years', '$7,919.34', '$1,468.63']
        ]

    def test_03_lowest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_LowestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '1.8 years', '$960.92', '$8.86'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.7 years', '$6,988.24', '$1,489.59'
            ],
            ['Totals', '$144.97', '4.7 years', '$7,949.15', '$1,498.44']
        ]

    def test_04_lowest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID,
                                      'table_LowestInterestRateFirstMethod'
                                      )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '1.8 years', '$960.92', '$8.86'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.7 years', '$6,988.24', '$1,489.59'
            ],
            ['Totals', '$144.97', '4.7 years', '$7,949.15', '$1,498.44']
        ]

    def test_05_min_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_MinPaymentMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '13.5 years', '$8,664.86', '$3,166.21'
            ],
            ['Totals', '$144.97', '13.5 years', '$9,627.86', '$3,177.15']
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestSettings(AcceptanceHelper):

    def test_000_setup_interest_charge_in_db(self, testdb):
        acct = testdb.query(Account).get(4)
        stmt = testdb.query(OFXStatement).get(7)
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='%s-MANUAL-CCPAYOFF' % dtnow().strftime('%Y%m%d%H%M%S'),
            trans_type='debit',
            date_posted=stmt.as_of,
            amount=Decimal('-46.9061'),
            name='Interest Charged - MANUALLY ENTERED',
            is_interest_charge=True
        )
        testdb.add(txn)
        testdb.commit()

    def test_00_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is None

    def test_01_no_config(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements(By.TAG_NAME, 'form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_onetime_frm_1'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            },
            'onetime': {
                1: {
                    'enabled': False,
                    'date': '',
                    'amount': ''
                }
            }
        }

    def test_02_input_and_save(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        assert 'disabled' in selenium.find_element(By.ID,
                                                   'btn_recalc_payoffs').get_attribute('class')
        set_payoff_form(selenium, 'increase', 1, True, '2017-05-14', '160.23')
        assert 'disabled' not in selenium.find_element(By.ID,
                                                       'btn_recalc_payoffs').get_attribute('class')
        selenium.find_element(By.LINK_TEXT, '(add another increase)').click()
        set_payoff_form(selenium, 'increase', 2, False, '2017-06-27', '-56.78')
        set_payoff_form(selenium, 'onetime', 1, False, '2017-07-21', '98.76')
        selenium.find_element(By.LINK_TEXT,
                              '(add another one-time additional payment)').click()
        set_payoff_form(selenium, 'onetime', 2, True, '2017-07-04', '54.32')
        selenium.find_element(By.ID, 'btn_recalc_payoffs').click()
        assert len(selenium.find_elements(By.CLASS_NAME, 'formfeedback')) == 0

    def test_03_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is not None
        assert b.value == json.dumps({
            'increases': [
                {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                },
                {
                    'enabled': False,
                    'date': '2017-06-27',
                    'amount': '-56.78'
                }
            ],
            'onetimes': [
                {
                    'enabled': True,
                    'date': '2017-07-04',
                    'amount': '54.32'
                },
                {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            ]
        }, sort_keys=True)

    def test_04_form_fill(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        form_ids = [
            s.get_attribute('id')
            for s in selenium.find_elements(By.TAG_NAME, 'form')
        ]
        assert sorted(form_ids) == [
            'min_payment_frm',
            'payoff_increase_frm_1',
            'payoff_increase_frm_2',
            'payoff_onetime_frm_1',
            'payoff_onetime_frm_2'
        ]
        assert get_payoff_forms(selenium) == {
            'increase': {
                1: {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                },
                2: {
                    'enabled': False,
                    'date': '2017-06-27',
                    'amount': '-56.78'
                }
            },
            'onetime': {
                1: {
                    'enabled': True,
                    'date': '2017-07-04',
                    'amount': '54.32'
                },
                2: {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            }
        }

    def test_05_highest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_HighestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$179.55',
                '4.0 years', '$6,734.42', '$1,235.77'
            ],
            ['Totals', '$214.55', '4.0 years', '$7,697.42', '$1,246.71']
        ]

    def test_06_highest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID,
                                      'table_HighestInterestRateFirstMethod'
                                      )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$179.55',
                '4.0 years', '$6,734.42', '$1,235.77'
            ],
            ['Totals', '$214.55', '4.0 years', '$7,697.42', '$1,246.71']
        ]

    def test_07_lowest_balance_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_LowestBalanceFirstMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$104.58',
                '1.3 years', '$958.10', '$6.04'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.1 years', '$6,809.61', '$1,310.96'
            ],
            ['Totals', '$214.55', '4.1 years', '$7,767.71', '$1,317.00']
        ]

    def test_08_lowest_interest_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID,
                                      'table_LowestInterestRateFirstMethod'
                                      )
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$104.58',
                '1.3 years', '$958.10', '$6.04'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '4.1 years', '$6,809.61', '$1,310.96'
            ],
            ['Totals', '$214.55', '4.1 years', '$7,767.71', '$1,317.00']
        ]

    def test_09_min_payment(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        table = selenium.find_element(By.ID, 'table_MinPaymentMethod')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                'CreditOne (3) ($952.06 @ 1.00%)', '$35.00',
                '2.3 years', '$963.00', '$10.94'
            ],
            [
                'CreditTwo (4) ($5,498.65 @ 10.00%)', '$109.97',
                '13.5 years', '$8,664.86', '$3,166.21'
            ],
            ['Totals', '$144.97', '13.5 years', '$9,627.86', '$3,177.15']
        ]

    def test_10_input_and_save_remove_lines(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/credit-payoff')
        selenium.find_element(By.ID, 'rm_increase_2_link').click()
        selenium.find_element(By.ID, 'rm_onetime_1_link').click()
        selenium.find_element(By.ID, 'btn_recalc_payoffs').click()
        assert len(selenium.find_elements(By.CLASS_NAME, 'formfeedback')) == 0

    def test_11_verify_db(self, testdb):
        b = testdb.query(DBSetting).get('credit-payoff')
        assert b is not None
        assert b.value == json.dumps({
            'increases': [
                {
                    'enabled': True,
                    'date': '2017-05-14',
                    'amount': '160.23'
                }
            ],
            'onetimes': [
                {
                    'enabled': False,
                    'date': '2017-07-21',
                    'amount': '98.76'
                }
            ]
        }, sort_keys=True)



"""Add Project and BoMItem models

Revision ID: ceb73ddf66eb
Revises: 8627a80874a6
Create Date: 2017-07-04 09:53:57.044088

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ceb73ddf66eb'
down_revision = '8627a80874a6'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'projects',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=40), nullable=True),
        sa.Column('notes', sa.String(length=254), nullable=True),
        sa.Column('is_active', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_projects')),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'bom_items',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=254), nullable=True),
        sa.Column('notes', sa.String(length=254), nullable=True),
        sa.Column('quantity', sa.Integer(), nullable=True),
        sa.Column(
            'unit_cost', sa.Numeric(precision=10, scale=4), nullable=True
        ),
        sa.Column('url', sa.String(length=254), nullable=True),
        sa.Column('is_active', sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ['project_id'], ['projects.id'],
            name=op.f('fk_bom_items_project_id_projects')
        ),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_bom_items')),
        mysql_engine='InnoDB'
    )


def downgrade():
    op.drop_table('bom_items')
    op.drop_table('projects')



from __future__ import absolute_import
from __future__ import unicode_literals
import hashlib
try:
    # python 3
    from io import StringIO, BytesIO
    IS_PYTHON_2 = False
except ImportError:
    # python 2
    from StringIO import StringIO
    IS_PYTHON_2 = True

from bs4 import BeautifulSoup
from ofxparse import OfxParser

from biweeklybudget.vendored.ofxclient.client import Client


class Institution(object):
    """Represents an institution or bank

    :param id: FI Id
    :type id: string
    :param org: FI Org
    :type org: string
    :param url: FI Url
    :type url: string
    :param username: Customer username or member id
    :type username: string
    :param password: Customer password or PIN
    :type password: string
    :param broker_id: FI Broker ID (optional)
    :type broker_id: string
    :param description: Description of the bank (optional)
    :type description: string or None
    :param client_args: :py:class:`ofxclient.Client` kwargs (optional)
    :type client_args: dict

    Values for many of the parameters need to come from some sort of
    OFX registry which knows about each banks particular setup.

    For help obtaining this sort of information; please see the
    :py:mod:`ofxhome` python module and/or the `OFX Home <http://ofxhome.com>`_
    website.

    Example::

      from biweeklybudget.vendored.ofxclient import Institution

      inst = Institution(
                id       = '3101',
                org      = 'AMEX',
                url      = 'https://online.americanexpress.com/myca\
                            /ofxdl/desktop/desktop Download.do?\
                            request_type=nl_ofxdownload',
                username = 'gene',
                password = 'wilder'
      )

      for a in inst.accounts():
          print(a.statement(days=5).balance)


    """
    def __init__(self, id, org, url, username, password,
                 broker_id='', description=None, client_args={}):
        self.id = id
        self.org = org
        self.url = url
        self.broker_id = broker_id
        self.username = username
        self.password = password
        self.description = description or self._default_description()
        self.client_args = client_args

    def client(self):
        """Build a :py:class:`ofxclient.Client` for talking with the bank

        It implicitly passes in the ``client_args`` that were passed
        when instantiating this ``Institution``.

        :rtype: :py:class:`ofxclient.Client`
        """
        return Client(institution=self, **self.client_args)

    def local_id(self):
        """Locally generated unique account identifier.

        :rtype: string
        """
        return hashlib.sha256(("%s%s" % (
            self.id,
            self.username)).encode()).hexdigest()

    def _default_description(self):
        return self.org

    def authenticate(self, username=None, password=None):
        """Test the authentication credentials

        Raises a ``ValueError`` if there is a problem authenticating
        with the human readable reason given by the institution.

        :param username: optional username (use self.username by default)
        :type username: string or None
        :param password: optional password (use self.password by default)
        :type password: string or None
        """

        u = self.username
        p = self.password
        if username and password:
            u = username
            p = password

        client = self.client()
        query = client.authenticated_query(username=u, password=p)
        res = client.post(query)
        ofx = BeautifulSoup(res, 'lxml')

        sonrs = ofx.find('sonrs')
        code = int(sonrs.find('code').contents[0].strip())

        try:
            status = sonrs.find('message').contents[0].strip()
        except Exception:
            status = ''

        if code == 0:
            return 1

        raise ValueError(status)

    def accounts(self):
        """Ask the bank for the known :py:class:`ofxclient.Account` list.

        :rtype: list of :py:class:`ofxclient.Account` objects
        """
        from biweeklybudget.vendored.ofxclient.account import Account
        client = self.client()
        query = client.account_list_query()
        resp = client.post(query)
        resp_handle = StringIO(resp)

        if IS_PYTHON_2:
            parsed = OfxParser.parse(resp_handle)
        else:
            parsed = OfxParser.parse(BytesIO(resp_handle.read().encode()))

        return [Account.from_ofxparse(a, institution=self)
                for a in parsed.accounts]

    def serialize(self):
        """Serialize predictably for use in configuration storage.

        Output looks like this::

          {
            'local_id':    'unique local identifier',
            'id':          'FI Id',
            'org':         'FI Org',
            'url':         'FI OFX Endpoint Url',
            'broker_id':   'FI Broker Id',
            'username':    'Customer username',
            'password':    'Customer password',
            'description': 'descr',
            'client_args': {
                'id':     'random client id - see Client() for default',
                'app_id': 'app name - see Client() for default',
                '...':    'see Client() for other options'
            }
          }

        :rtype: nested dictionary
        """
        return {
            'id': self.id,
            'org': self.org,
            'url': self.url,
            'broker_id': self.broker_id,
            'username': self.username,
            'password': self.password,
            'description': self.description,
            'client_args': self.client().init_args,
            'local_id': self.local_id()
        }

    @staticmethod
    def deserialize(raw):
        """Instantiate :py:class:`ofxclient.Institution` from dictionary

        :param raw: serialized ``Institution``
        :param type: dict per :py:method:`~Institution.serialize`
        :rtype: subclass of :py:class:`ofxclient.Institution`
        """
        return Institution(
            id=raw['id'],
            org=raw['org'],
            url=raw['url'],
            broker_id=raw.get('broker_id', ''),
            username=raw['username'],
            password=raw['password'],
            description=raw.get('description', None),
            client_args=raw.get('client_args', {})
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import hvac
import logging

logger = logging.getLogger(__name__)


class SecretMissingException(Exception):

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return repr(self.path)


class Vault(object):
    """
    Provides simpler access to Vault
    """

    def __init__(self, addr=None, token_path=None):
        """
        Connect to Vault and maintain connection

        :param addr: URL to connect to Vault at. If None, defaults to
          :py:attr:`biweeklybudget.settings.VAULT_ADDR`.
        :type addr: str
        :param token_path: path to read Vault token from. If None, defaults to
          :py:attr:`biweeklybudget.settings.TOKEN_PATH`.
        :type token_path: str
        """
        if addr is None and 'VAULT_ADDR' in os.environ:
            addr = os.environ['VAULT_ADDR']
        if token_path is None and 'TOKEN_PATH' in os.environ:
            token_path = os.environ['TOKEN_PATH']
        # if not in constructor or environment, use settings if possible
        if addr is None or token_path is None:
            try:
                from biweeklybudget import settings
                if addr is None:
                    addr = settings.VAULT_ADDR
                if token_path is None:
                    token_path = settings.TOKEN_PATH
            except Exception:
                logger.error('ERROR: you must either set the SETTINGS_MODULE '
                             'environment variable to use a settings module, '
                             'or export VAULT_ADDR and TOKEN_PATH environment '
                             'variables.')
                raise SystemExit(1)
        token_path = os.path.expanduser(token_path)
        logger.debug('Connecting to Vault at %s with token from %s',
                     addr, token_path)
        with open(token_path, 'r') as fh:
            tkn = fh.read().strip()
        self.conn = hvac.Client(url=addr, token=tkn)
        assert self.conn.is_authenticated()
        logger.debug('Connected to Vault')

    def read(self, secret_path):
        """
        Read and return a secret from Vault. Return only the data portion.

        :param secret_path: path to read in Vault
        :type secret_path: str
        :return: secret data
        :rtype: dict
        """
        res = self.conn.read(secret_path)
        if res is None:
            raise SecretMissingException(secret_path)
        return res['data']



"""txnreconcile unique constraints

Revision ID: f136418fc4ab
Revises: a8f1228af250
Create Date: 2017-04-25 17:23:48.322303

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f136418fc4ab'
down_revision = 'a8f1228af250'
branch_labels = None
depends_on = None


def upgrade():
    op.create_unique_constraint(
        op.f('uq_txn_reconciles_ofx_account_id'),
        'txn_reconciles',
        ['ofx_account_id', 'ofx_fitid']
    )
    op.create_unique_constraint(
        op.f('uq_txn_reconciles_txn_id'),
        'txn_reconciles',
        ['txn_id']
    )


def downgrade():
    op.drop_constraint(
        op.f('uq_txn_reconciles_txn_id'),
        'txn_reconciles',
        type_='unique'
    )
    op.drop_constraint(
        op.f('uq_txn_reconciles_ofx_account_id'),
        'txn_reconciles',
        type_='unique'
    )
